"""
    Policy decision point implementation
"""

from copy import deepcopy

from .call_api import CallAPI
from .policy import PolicySchema, AccessType
from .policy.utils import Utils
from .result_access import ResultAccess


class PolicyDecisionPoint(object):
    """
        Policy decision point
    """

    class AccountType:
        SYSTEM = "system"
        NORMAL = "normal"

    class AccessLevel:
        LIST = "list"
        READ = "read"
        ADD = "add"
        EDIT = "edit"
        DELETE = "delete"
        OTHER = "other"

    def __init__(self,
                 merchant_id: str, resource: str, action: str, account_id: str = None, user_info: dict = None,
                 data_before: dict = None, data_after: dict = None, environment: dict = None,
                 account_type=AccountType.NORMAL):
        if not merchant_id:
            raise ValueError("merchant_id required")
        if not resource:
            raise ValueError("resource required")
        if not action:
            raise ValueError("action required")

        if account_type == PolicyDecisionPoint.AccountType.NORMAL:
            if not account_id:
                raise ValueError("account normal account_id required")
            if not user_info:
                user_info = CallAPI.admin_get_account_info(merchant_id, account_id)

        self.account_type = account_type
        self.merchant_id = merchant_id
        self.account_id = account_id
        self.resource = resource
        self.action = action
        self.request_access = {}
        self.result_access = ResultAccess()
        self.request_access.update({
            "user": user_info if user_info else {},
            "env": self.get_info_environment(environment),
            self.resource: data_before if data_before else {}
        })
        self.data_after = data_after if data_after else {}
        self.data_before = data_before if data_before else {}

    def get_policy_statement_for_target(self):
        return CallAPI.admin_get_list_statement(self.merchant_id, self.account_id, self.resource, self.action)

    def is_allowed(self):
        """
            Check if authorization request is allowed

            :param request: request object
            :return: True if authorized else False
        """
        try:
            if self.account_type == PolicyDecisionPoint.AccountType.SYSTEM:
                self.result_access.set_allow_access(True)
                return self.result_access
            list_statement = self.get_policy_statement_for_target()
            if list_statement:
                action_info = CallAPI.admin_get_json_action(merchant_id=self.merchant_id)
                if not action_info:
                    raise ValueError("action for merchant_id {} not found".format(self.merchant_id))
                if not action_info.get(self.action):
                    raise ValueError("action {} not found".format(self.action))
                access_level = action_info.get(self.action, {}).get("access_level").lower()

                statement_check_allow, statement_check_deny = self.get_statement_by_type(list_statement, access_level)
                if statement_check_deny:
                    if self.check_list_statement_is_deny(statement_check_deny):
                        self.result_access.set_allow_access(False)
                        return self.result_access
                if statement_check_allow:
                    self.result_access.set_allow_access(self.check_list_statement_is_allow(statement_check_allow))
                else:
                    self.result_access.set_allow_access(False)
            else:
                self.result_access.set_allow_access(False)
        except Exception as err:
            print("Exception: {}".format(err))
            self.result_access.set_allow_access(False)

        return self.result_access

    def get_info_environment(self, environment):
        if not environment:
            environment = Utils.get_info_from_header_request()
            if environment:
                if environment.get("env:user_agent"):
                    environment.update(Utils.parse_user_agent(environment.get("env:user_agent")))
        environment = environment if environment else {}
        environment.update({"env:current_time": Utils.get_date_utcnow()})
        return environment

    def get_value_from_variable(self, str_variable):
        variables = Utils.get_field_key_from_variable(str_variable)
        if variables:
            if len(variables) == 1 and Utils.check_field_is_variable(variables[0]):
                field_value = self.get_value_from_field(variables[0])
                return field_value
            for field_key in variables:
                field_value = self.get_value_from_field(field_key)
                if field_value is not None:
                    field_value = str(field_value)
                    str_variable = Utils.replace_variable_to_value(field_key, field_value, str_variable)
        return str_variable

    def get_value_from_field(self, field_key):
        resource_name, resource_key = Utils.split_delemiter_resource(field_key)
        if resource_name and resource_key:
            if self.request_access.get(resource_name) and isinstance(self.request_access.get(resource_name), dict):
                data_resource = self.request_access.get(resource_name)
                value = Utils.get_nested_value(data_resource, resource_key)
                return value
        return None

    def get_statement_by_type(self, list_statement: list, access_level: str):
        statement_check_allow, statement_check_deny = [], []
        for statement in list_statement:
            statement_copy = deepcopy(statement)
            fields = []
            if statement_copy.get("resource_field") and statement_copy.get("resource_field").get(self.resource):
                fields = statement_copy.get("resource_field").get(self.resource)
            if fields:
                statement_copy["fields"] = fields
            if access_level == PolicyDecisionPoint.AccessLevel.LIST:
                condition_check = []
                condition_filter = []
                for item_cond in statement_copy.get("condition"):
                    if item_cond.get("field").startswith("user") or item_cond.get("field").startswith("env"):
                        condition_check.append(item_cond)
                    else:
                        values = [self.get_value_from_variable(i) for i in item_cond.get("values")]
                        item_cond["values"] = values
                        condition_filter.append(item_cond)
                if statement_copy.get("effect") == AccessType.ALLOW_ACCESS:
                    statement_check_allow.append({
                        **statement_copy,
                        "condition": condition_check
                    })
                else:
                    statement_check_deny.append({
                        **statement_copy,
                        "condition": condition_check
                    })
                if condition_filter:
                    self.result_access.add_filter_config(
                        {"effect": statement_copy.get("effect"), "condition": condition_filter})
            elif access_level == PolicyDecisionPoint.AccessLevel.READ:
                if statement_copy.get("effect") == AccessType.ALLOW_ACCESS:
                    statement_check_allow.append(statement_copy)
                else:
                    statement_check_deny.append(statement_copy)
                if fields:
                    self.result_access.add_display_config({
                        "effect": statement_copy.get("effect"),
                        "fields": fields,
                    })
            else:
                if access_level == PolicyDecisionPoint.AccessLevel.ADD:
                    self.data_after.update(self.data_before)
                if fields:
                    statement_copy["check_field"] = 1
                if statement_copy.get("effect") == AccessType.ALLOW_ACCESS:
                    statement_check_allow.append(statement_copy)
                else:
                    statement_check_deny.append(statement_copy)
        return statement_check_allow, statement_check_deny

    def check_list_statement_is_deny(self, list_statement: list):
        result_deny = False
        try:
            for statement in list_statement:
                statement["request_access"] = self.request_access
                policy_schema = PolicySchema().load(statement)
                if policy_schema.is_allowed():
                    if statement.get("check_field") and statement.get("fields"):
                        if not Utils.field_in_body_request(statement.get("fields"), self.data_after):
                            continue
                    result_deny = True
                    break
        except Exception as err:
            print("check_list_statement_is_deny err: {}".format(err))
            result_deny = False
        return result_deny

    def check_list_statement_is_allow(self, list_statement: list):
        result_allow = True
        try:
            for statement in list_statement:
                # đối với action là edit create mà resource liên quan đến field thì kiểm tra field có trong body
                # yêu cầu truy cập hay ko
                # case: policy cho sửa 3 field, dữ liệu gửi lên có 5 field trong đó có 1 field trong policy
                # thì hiện tại chưa chặn dc do có policy * all, sau này bỏ policy * all thì
                # sửa lại kiểm tra danh sách field gửi lên phải nằm trong policy
                # hiện tại chỉ cần có 1 field trong dữ liệu gửi lên có trong policy là dc cho qua

                statement["request_access"] = self.request_access
                policy_schema = PolicySchema().load(statement)
                if not policy_schema.is_allowed():
                    result_allow = False
                    break
                if statement.get("check_field") and statement.get("fields"):
                    if not Utils.field_in_body_request(statement.get("fields"), self.data_after):
                        result_allow = False
                        break

        except Exception as err:
            print("check_list_statement_is_allow err: {}".format(err))
            result_allow = False
        return result_allow
